/*
 * C
 *
 * Copyright 2012-2013 IS2T. All rights reserved.
 * For demonstration purpose only.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#include <stdlib.h>
#include <stdio.h>
#include "LLMJVM_impl.h"
#include "cpu.h"
#if MCU == 2
#include "stm32f2xx_rcc.h"
#elif MCU == 4
#include "stm32f4xx_rcc.h"
#else
#error Invalid MCU
#endif

#define MEM_CODE_START (0)
#define MEM_CODE_END (0x20000000)

//Frequency of the system tick
#define MICROJVM_TICK_FREQUENCY (100) // a tick every 10 ms

//Convert a time in ms into a number of ticks
#define timeToTick(time) ((time * MICROJVM_TICK_FREQUENCY) / 1000)

//Global tick counter
static int64_t microjvm_tick = 0;

//Absolute time in ticks at which timer will be launched
static int64_t microjvm_nextWakeupTick = INT64_MAX;

//Set to 1 if the VM must be woken up
static volatile uint8_t microjvm_wakeup = 0;

//Offset in ms from system time to application time
int64_t applicationTimeOffset = 0;

// Prototypes
int64_t LLMJVM_IMPL_timeToTick(int64_t time);
int64_t LLMJVM_IMPL_tickToTime(int64_t tick);


// Implementations

int32_t LLMJVM_IMPL_initialize()
{
	RCC_ClocksTypeDef RCC_Clocks;

	// Configure SysTick
	RCC_GetClocksFreq(&RCC_Clocks);

	// Check configuration was done correctly
	if( SysTick_Config(RCC_Clocks.HCLK_Frequency / MICROJVM_TICK_FREQUENCY) == 0)
	{
		return LLMJVM_OK;
	}
	else
	{
		// The parameter is incorrect (value too big)
		puts("SYSTICK not initialized");
		return LLMJVM_ERROR;
	}
}

void SysTick_Handler(void)
{
	if((++microjvm_tick) >= microjvm_nextWakeupTick){
		LLMJVM_schedule();
	}
}

int32_t LLMJVM_IMPL_vmTaskStarted(){
	return LLMJVM_OK;
}

int32_t LLMJVM_IMPL_scheduleRequest(int64_t absoluteTime){
	int64_t absoluteTick = timeToTick(absoluteTime);

	//First check if absolute time is lower than current schedule time
	if(absoluteTick < microjvm_nextWakeupTick){
		microjvm_nextWakeupTick = absoluteTick;

		//Checks if absoluteTick has already been reached	
		if(absoluteTick <= microjvm_tick){
			//absoluteTime has been reached, notify the vm now
			return LLMJVM_schedule();
		}
	}

	return LLMJVM_OK;
}

int32_t LLMJVM_IMPL_idleVM(){
	//loop until wakeup time is reached
	while(!microjvm_wakeup);

	return LLMJVM_OK; //always OK, the MicroJvm virtual machine cannot be interrupted
}

int32_t LLMJVM_IMPL_wakeupVM(){
	microjvm_wakeup = 1;
	microjvm_nextWakeupTick = INT64_MAX;//reset next wakeup time
	return LLMJVM_OK;
}

int32_t LLMJVM_IMPL_ackWakeup(){
	microjvm_wakeup = 0;
	return LLMJVM_OK;
}

/**
 * <p>
 * Returns the ID of the current task.
 * </p>
 */
int32_t LLMJVM_IMPL_getCurrentTaskID(){
	return 1;
}


void LLMJVM_IMPL_setApplicationTime(int64_t t){
	int64_t currentTime = LLMJVM_IMPL_getCurrentTime(1);
	applicationTimeOffset = t - currentTime;
}



int64_t LLMJVM_IMPL_getCurrentTime(uint8_t system){
	int64_t systemTime = LLMJVM_IMPL_tickToTime(microjvm_tick);
	int64_t timeFromPreviousTick = ( (SysTick->LOAD - SysTick->VAL) * 1000) / (SysTick->LOAD * MICROJVM_TICK_FREQUENCY);

	systemTime = systemTime + timeFromPreviousTick;

	if(system == 1)
		return systemTime;
	else
		return systemTime + applicationTimeOffset;
}

int64_t LLMJVM_IMPL_getTimeNanos(){
	return LLMJVM_IMPL_getCurrentTime(1) * 1000000;
}

uint8_t LLMJVM_IMPL_isInReadOnlyMemory(void* ptr){
	if(
#if MEM_CODE_START != 0
			// avoid warning #186-D: pointless comparison of unsigned integer with zero
			(uint32_t)ptr >= MEM_CODE_START &&
#endif
			(uint32_t)ptr < MEM_CODE_END
	)
		return 1;
	else
		return 0;
}


/*
 * Convert a time in milliseconds in ticks.
 * Result is round up to the next tick (ie. converting back the resulting ticks
 * to milliseconds results in a value greather than or equals to given time).
 */
int64_t LLMJVM_IMPL_timeToTick(int64_t time){
	time = time * MICROJVM_TICK_FREQUENCY;
	return (time + (1000-1)) / 1000;
}

/*
 * Convert a number of ticks in time in milliseconds.
 */
int64_t LLMJVM_IMPL_tickToTime(int64_t tick){
	return (tick * 1000) / MICROJVM_TICK_FREQUENCY;
}

int32_t LLMJVM_IMPL_shutdown(void)
{
	// nothing to do
	return LLMJVM_OK;
}

